#ifndef PARSER_H
#define PARSER_H

#include "vektor.h"
#include "monomial.h"
#include "polynomial.h"
#include "field.h"

class Parser
{
 public:
  virtual int parseInt()=0;
  virtual Monomial parseMonomial(int dimension)=0;
  virtual IntegerVector parseIntegerVector()=0;
};


class CharacterBasedParser : public Parser
{
 private:
  bool isVariable(int c);
  bool isDigit(int c);
  int variableIndex(int c);
 protected:
  virtual int getChar()=0;
  virtual void ungetChar(int c)=0;
 public:
  int nextNonBlank();
  int nextNonBlankDoNotGet();
  bool isLeftBrackets(int c);
  bool isRightBrackets(int c);
  int parseChar();
  int parseInt();
  double parseFloat();
  FieldElement parseFieldElement();
  FieldElement parseFieldElementFromInteger();
  Monomial parseMonomial(int dimension=-1);
  IntegerVector parseIntegerVector();
  IntegerVectorList parseIntegerVectorList();
  Term parseTerm(int dimension=-1);
  Polynomial parsePolynomial(int dimension=-1);
  PolynomialSet parsePolynomialSet(int dimension=-1);
  PolynomialSetList parsePolynomialSetList();
};


class FileParser : public CharacterBasedParser
{
  FILE *f;
 protected:
  virtual int getChar();
  virtual void ungetChar(int c);
 public:
  FileParser(FILE *f);
};


class StringParser : public CharacterBasedParser
{
  char *s;
  int index;
  bool hasUngotten;
  char ungotten;
 protected:
  virtual int getChar();
  virtual void ungetChar(int c);
 public:
  StringParser(char *s);
};

#endif
